from __future__ import annotations

from typing import Literal

from nsidc.iceflow.data.models import (
    ATM1BDataset,
    ATM1BShortName,
    Dataset,
    DatasetShortName,
)


class ILATM1BDataset(ATM1BDataset):
    short_name: ATM1BShortName = "ILATM1B"
    version: Literal["1", "2"]


class BLATM1BDataset(ATM1BDataset):
    short_name: ATM1BShortName = "BLATM1B"
    # There is only 1 version of BLATM1B
    version: Literal["1"] = "1"


class ILVIS2Dataset(Dataset):
    short_name: DatasetShortName = "ILVIS2"
    version: Literal["1", "2"]


class GLAH06Dataset(Dataset):
    short_name: DatasetShortName = "GLAH06"
    # Note: some dataset versions are padded with zeros like GLAH06. NSIDC
    # documentation refers to "version 34", but CMR only recognizes "034".  As a
    # rule-of-thumb, ICESat-2, SMAP, and GLAH/GLA datasets have zero padding.
    version: Literal["034"] = "034"


ALL_SUPPORTED_DATASETS: list[Dataset] = [
    ILATM1BDataset(version="1"),
    ILATM1BDataset(version="2"),
    BLATM1BDataset(version="1"),
    ILVIS2Dataset(version="1"),
    ILVIS2Dataset(version="2"),
    GLAH06Dataset(),
]



from __future__ import annotations

import datetime as dt
import importlib.metadata
from typing import Any

project = "nsidc-iceflow"
copyright = f"{dt.date.today().year}, NSIDC"
author = "NSIDC"
version = release = importlib.metadata.version("nsidc-iceflow")

extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

source_suffix = [".rst", ".md"]
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

html_theme = "furo"

html_theme_options: dict[str, Any] = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/NSIDC/nsidc-iceflow",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    "source_repository": "https://github.com/NSIDC/nsidc-iceflow",
    "source_branch": "main",
    "source_directory": "docs/",
}

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}


autodoc_mock_imports = [
    "pydantic*",
    "pandera*",
    "pandera.api",
]

nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
    ("py:class", "pandera.typing.pandas.DataFrame"),
    ("py:class", "pydantic.main.BaseModel"),
    ("py:class", "pandera.api.pandas.model_config.BaseConfig"),
    ("py:class", "pandera.api.pandas.model.DataFrameModel"),
    ("py:class", "pandera.api.dataframe.model.Config"),
    # avoids warning:
    # `iceflow/docs/api/nsidc.iceflow.data.rst:13: WARNING: py:class reference target not found: data [ref.class]`
    ("py:class", "data"),
]

always_document_param_types = True

nb_execution_mode = "off"


def copy_notebook_images(app, _exception):
    """Copy notebook static images into place post-build."""
    import shutil
    from pathlib import Path

    source_dir = Path(app.srcdir) / "img"
    dest_dir = Path(app.outdir) / "img"

    if dest_dir.exists():
        shutil.rmtree(dest_dir)

    shutil.copytree(source_dir, dest_dir)


def setup(app):
    """Setup build-finished action to copy images into place.

    Because the images are only referenced in jupyter notebooks, and not source
    files parsed directly by sphinx, the images will not be copied on their own.
    """
    app.connect("build-finished", copy_notebook_images)



"""
Copyright (c) 2025 NSIDC. All rights reserved.

iceflow: Harmonized access to (pre)OIB/IceSAT/IceSAT2 data

Users interact with `nsidc-iceflow` by:

* Searching for data that match an area of interest/time (`find_iceflow_data`)
* Downloading data (`download_iceflow_results`)
* (Optional) Creating a parquet datastore to facilitate reading the data (`make_iceflow_parquet`)
* Reading and doing analysis with the data (`dask.dataframe.read_parquet`,
  `read_iceflow_datafiles`)
* (Optional, if using `read_iceflow_datafiles`) Transform the lat/lon/elev data
  into a target International Terrestrial Reference Frame (ITRF) (`transform_itrf`)

"""

from __future__ import annotations

from nsidc.iceflow.api import make_iceflow_parquet
from nsidc.iceflow.data.fetch import download_iceflow_results, find_iceflow_data
from nsidc.iceflow.data.models import (
    Dataset,
    IceflowDataFrame,
)
from nsidc.iceflow.data.read import read_iceflow_datafiles
from nsidc.iceflow.data.supported_datasets import (
    ALL_SUPPORTED_DATASETS,
    BLATM1BDataset,
    GLAH06Dataset,
    ILATM1BDataset,
    ILVIS2Dataset,
)
from nsidc.iceflow.itrf.converter import transform_itrf

__version__ = "v1.1.0"


__all__ = [
    "ALL_SUPPORTED_DATASETS",
    "BLATM1BDataset",
    "Dataset",
    "GLAH06Dataset",
    "ILATM1BDataset",
    "ILVIS2Dataset",
    "IceflowDataFrame",
    "__version__",
    "download_iceflow_results",
    "find_iceflow_data",
    "make_iceflow_parquet",
    "read_iceflow_datafiles",
    "transform_itrf",
]



"""Add format and test to collections."""

from __future__ import annotations

from invoke import Collection

from . import docs, test

ns = Collection()
ns.add_collection(test)
ns.add_collection(docs)



from __future__ import annotations

from invoke import task

from .util import PROJECT_DIR, print_and_run


def _build_apidocs_cmd():
    return (
        f"sphinx-apidoc -o {PROJECT_DIR}/docs/api/ --no-toc"
        f" --module-first --implicit-namespaces --force {PROJECT_DIR}/src/nsidc"
    )


@task()
def clean(_ctx):
    print_and_run(f"rm -rf {PROJECT_DIR}/docs/api/*")
    print_and_run(f"rm -rf {PROJECT_DIR}/docs/_build/*")


@task(pre=[clean])
def build(_ctx):
    """Build docs."""
    # (re)generate the api docs
    print_and_run(
        _build_apidocs_cmd(),
        pty=True,
    )

    # Build the docs
    print_and_run(
        (
            "sphinx-build --keep-going -n -T -b=html"
            f" {PROJECT_DIR}/docs {PROJECT_DIR}/docs/_build/html/"
        ),
        pty=True,
    )


@task(pre=[clean])
def watch(_ctx):
    print_and_run(
        (
            f'sphinx-autobuild --pre-build "{_build_apidocs_cmd()}"'
            f" --watch {PROJECT_DIR}/src"
            f" --watch {PROJECT_DIR}/docs/notebooks"
            f" --ignore {PROJECT_DIR}/docs/notebooks/.ipynb_checkpoints"
            f" --ignore {PROJECT_DIR}/docs/notebooks/_sources"
            f" --ignore {PROJECT_DIR}/docs/notebooks/_static"
            f" --ignore {PROJECT_DIR}/docs/notebooks/downloaded-data/"
            f" {PROJECT_DIR}/docs {PROJECT_DIR}/docs/_build/html"
        ),
        pty=True,
    )



from __future__ import annotations

import re

ITRF_REGEX = re.compile(r"^ITRF\d{2}(\d{2})?$")


def check_itrf(itrf_str: str) -> bool:
    """Check if the given string is  a valid ITRF.

    Based on a regex match. ITRF strings are conventionally "ITRF" followed by
    the 2-digit or 4-digit year (e.g., "ITRF93" or "ITRF2008").

    ITRFs prior to the 2000s conventionally used a 2-digit year (e.g.,
    "ITRF88").

    ITRF2020 is recognized as "ITRF20".
    """

    match = ITRF_REGEX.match(itrf_str)

    return match is not None



"""Task to run tests for this package."""

from __future__ import annotations

from invoke import task

from .util import PROJECT_DIR, print_and_run


@task(aliases=["mypy"])
def typecheck(_ctx):
    """Run mypy typechecking."""
    print_and_run(
        "mypy",
        pty=True,
    )

    print("🎉🦆 Type checking passed.")


@task()
def unit(_ctx):
    """Run all unit tests with pytest."""
    print_and_run(
        f"PYTHONPATH={PROJECT_DIR}/src:$PYTHONPATH pytest tests/unit/ --capture=no",
        pty=True,
    )


@task()
def integration(_ctx):
    """Run integration tests with pytest."""
    print_and_run(
        f"PYTHONPATH={PROJECT_DIR}/src:$PYTHONPATH pytest tests/integration --capture=no",
        pty=True,
    )


@task()
def pytest(_ctx):
    """Run all tests with pytest."""
    print_and_run(
        f"PYTHONPATH={PROJECT_DIR}/src:$PYTHONPATH pytest tests/ --capture=no",
        pty=True,
    )


@task(pre=[typecheck, unit])
def ci(_ctx):
    """Run all CI tests"""


@task(
    pre=[
        typecheck,
        pytest,
    ],
    default=True,
)
def all(_ctx):
    """Run all of the tests."""



from __future__ import annotations

from nsidc.iceflow.itrf import check_itrf


def test_check_itrf():
    # These should return False
    assert not check_itrf("Not an ITRF string")
    assert not check_itrf("ITRF")

    # These should return True
    assert check_itrf("ITRF2008")
    assert check_itrf("ITRF88")



from __future__ import annotations

import datetime as dt
import re
from collections import namedtuple
from pathlib import Path
from typing import Literal

import numpy as np
import pandas as pd
import pandera as pa

from nsidc.iceflow.data.models import ILVIS2DataFrame

ILVIS2_COORDINATE_SETS = Literal["low_mode", "high_mode", "centroid", "highest_signal"]
ILVIS2_COORDINATE_SET_MAPPING: dict[ILVIS2_COORDINATE_SETS, dict[str, str]] = {
    # The center of the lowest detected mode within the waveform. Available in
    # both v1 and v2.
    "low_mode": {
        "latitude": "GLAT",
        "longitude": "GLON",
        "elevation": "ZG",
    },
    # The center of the highest detected mode within the waveform. Available in
    # both v1 and v2.
    "high_mode": {
        "latitude": "HLAT",
        "longitude": "HLON",
        "elevation": "ZH",
    },
    # Centroid of the corresponding LVIS Level-1B waveform. v1 only.
    "centroid": {
        "latitude": "CLAT",
        "longitude": "CLON",
        "elevation": "ZC",
    },
    # Highest detected signal. v2 only.
    "highest_signal": {
        "latitude": "TLAT",
        "longitude": "TLON",
        "elevation": "ZT",
    },
}
# In the valkyrie service that this code is based on, only the low_mode was
# exposed to users. This default replicates that behavior.
ILVIS2_DEFAULT_COORDINATE_SET: ILVIS2_COORDINATE_SETS = "low_mode"

Field = namedtuple("Field", ["name", "type", "scale_factor"])

"""
See: https://lvis.gsfc.nasa.gov/Data/Data_Structure/DataStructure_LDS104.html

Note: The LVIS site (above) and NSIDC data files use different names
for the fields. The list of field tuples below matches the
documentation on the LVIS site. This is done to simplify the code and
ease the mental mapping of the v1.0.4 and v2.0.2b fields to the
database.  The mapping between the names used below (same as LVIS
docs) and the field names used in the NSIDC files is:


  CLON = LONGITUDE_CENTROID
  CLAT = LATITUDE_CENTROID
  ZC = ELEVATION_CENTROID
  GLON = LONGITUDE_LOW
  GLAT = LATITUDE_LOW
  ZG = ELEVATION_LOW
  HLON = LONGITUDE_HIGH
  HLAT = LATITUDE_HIGH
  ZH = ELEVATION_HIGH

"""
ILVIS2_V104_FIELDS = [
    Field("LFID", None, np.uint64),
    Field("SHOTNUMBER", None, np.uint64),
    Field("TIME", 10**6, np.uint64),
    Field("CLON", 10**6, np.int64),
    Field("CLAT", 10**6, np.int64),
    Field("ZC", 10**6, np.int64),
    Field("GLON", None, np.float64),
    Field("GLAT", None, np.float64),
    Field("ZG", None, np.float64),
    Field("HLON", 10**6, np.int64),
    Field("HLAT", 10**6, np.int64),
    Field("ZH", 10**6, np.int64),
]

"""
See: https://lvis.gsfc.nasa.gov/Data/Data_Structure/DataStructure_LDS202.html
Note: Version 2.0.2b was used for Greenland 2017
"""
ILVIS2_V202b_FIELDS = [
    Field("LFID", None, np.uint64),
    Field("SHOTNUMBER", None, np.uint64),
    Field("TIME", 10**6, np.uint64),
    Field("GLON", None, np.float64),
    Field("GLAT", None, np.float64),
    Field("ZG", None, np.float64),
    Field("HLON", 10**6, np.int64),
    Field("HLAT", 10**6, np.int64),
    Field("ZH", 10**6, np.int64),
    Field("TLON", 10**6, np.int64),
    Field("TLAT", 10**6, np.int64),
    Field("ZT", 10**6, np.int64),
    Field("RH10", 10**3, np.int64),
    Field("RH15", 10**3, np.int64),
    Field("RH20", 10**3, np.int64),
    Field("RH25", 10**3, np.int64),
    Field("RH30", 10**3, np.int64),
    Field("RH35", 10**3, np.int64),
    Field("RH40", 10**3, np.int64),
    Field("RH45", 10**3, np.int64),
    Field("RH50", 10**3, np.int64),
    Field("RH55", 10**3, np.int64),
    Field("RH60", 10**3, np.int64),
    Field("RH65", 10**3, np.int64),
    Field("RH70", 10**3, np.int64),
    Field("RH75", 10**3, np.int64),
    Field("RH80", 10**3, np.int64),
    Field("RH85", 10**3, np.int64),
    Field("RH90", 10**3, np.int64),
    Field("RH95", 10**3, np.int64),
    Field("RH96", 10**3, np.int64),
    Field("RH97", 10**3, np.int64),
    Field("RH98", 10**3, np.int64),
    Field("RH99", 10**3, np.int64),
    Field("RH100", 10**3, np.int64),
    Field("AZIMUTH", 10**3, np.int64),
    Field("INCIDENT_ANGLE", 10**3, np.int64),
    Field("RANGE", 10**3, np.int64),
    Field("COMPLEXITY", 10**3, np.int64),
    Field("CHANNEL_ZT", None, np.uint8),
    Field("CHANNEL_ZG", None, np.uint8),
    Field("CHANNEL_RH", None, np.uint8),
]

"""Names of fields that contain longitude values. The values in these
fields will be shifted to the range [-180,180)."""
ILVIS2_LONGITUDE_FIELD_NAMES = ["CLON", "GLON", "HLON", "TLON"]


def _file_date(filename: str) -> dt.date:
    """Return the datetime from the ILVIS2 filename."""
    return dt.datetime.strptime(filename[9:18], "%Y_%m%d").date()


def _shift_lon(lon):
    """Shift longitude values from [0,360] to [-180,180]"""
    if lon >= 180.0:
        return lon - 360.0
    return lon


def _add_utc_datetime(df: pd.DataFrame, file_date) -> pd.DataFrame:
    """Add a `utc_datetime` column to the DataFrame, with values
    calculated from the given date and the `TIME` values in the
    dataset (seconds of the day).
    """
    df["utc_datetime"] = pd.to_datetime(file_date)
    df["utc_datetime"] = df["utc_datetime"] + pd.to_timedelta(df["TIME"], unit="s")

    return df


def _scale_and_convert(df: pd.DataFrame, fields) -> pd.DataFrame:
    """For any column in the list of Field named tuples, optionally scale
    the corresponding column in the DataFrame and convert the column
    type.
    """
    for name, scale_factor, dtype in fields:
        if scale_factor is not None:
            df.loc[:, name] *= scale_factor
        if dtype != df.dtypes[name]:
            df[name] = df[name].astype(dtype)

    return df


def _ilvis2_data(filepath: Path, file_date: dt.date, fields) -> pd.DataFrame:
    """Return an ILVIS2 file DataFrame, performing all necessary
    conversions / augmentation on the data.
    """
    field_names = [name for name, _, _ in fields]
    df = pd.read_csv(filepath, sep=r"\s+", comment="#", names=field_names)

    for col in ILVIS2_LONGITUDE_FIELD_NAMES:
        if col in df.columns:
            df[col] = df[col].apply(_shift_lon)

    df = _add_utc_datetime(df, file_date)

    df = _scale_and_convert(df, fields)

    return df


@pa.check_types()
def ilvis2_data(
    filepath: Path,
    coordinate_set: ILVIS2_COORDINATE_SETS = ILVIS2_DEFAULT_COORDINATE_SET,
) -> ILVIS2DataFrame:
    """Return the ilvis2 data given a filepath.

    Parameters
    ----------
    fn
        The filename (str) to read. This can be a file in the LVIS2
        v1.0.4 or v2.0.2b format.
        https://lvis.gsfc.nasa.gov/Data/Data_Structure/DataStructure_LDS104.html
        https://lvis.gsfc.nasa.gov/Data/Data_Structure/DataStructure_LDS202.html

    Returns
    -------
    data
        The ilvis2 (pandas.DataFrame) data.

    """
    filename = filepath.name
    match = re.search(r"_\D{2}(\d{4})_", filename)
    if not match:
        err = f"Failed to recognize {filename} as ILVIS2 data."
        raise RuntimeError(err)

    year = int(match.group(1))

    if year < 2017:
        # This corresponds to ILVIS v1
        dataset = "ILVIS2v1"
        the_fields = ILVIS2_V104_FIELDS
        # The user guide indicates ILVIS2 v1 data uses ITRF2000 as a reference frame:
        # https://nsidc.org/sites/default/files/documents/user-guide/ilvis2-v001-userguide.pdf
        itrf_str = "ITRF2000"
        # Ensure that the highest_signal coordinate set has not been selected.
        if coordinate_set == "highest_signal":
            raise ValueError(
                "ILVIS coordinate set 'highest_signal' is only available in v2 data."
            )
    else:
        # This corresponds to ILVIS v2
        dataset = "ILVIS2v2"
        the_fields = ILVIS2_V202b_FIELDS
        # The user guide indicates ILVIS2 v1 data uses ITRF2008 as a reference frame:
        # https://nsidc.org/sites/default/files/documents/user-guide/ilvis2-v002-userguide.pdf
        itrf_str = "ITRF2008"
        # Ensure that the centroid coordinate set has not been selected.
        if coordinate_set == "centroid":
            raise ValueError(
                "ILVIS coordinate set 'centroid' is only available in v1 data."
            )

    file_date = _file_date(filename)

    data = _ilvis2_data(filepath, file_date, the_fields)
    data["ITRF"] = itrf_str
    data["dataset"] = dataset

    coordinate_fields = ILVIS2_COORDINATE_SET_MAPPING[coordinate_set]
    data["latitude"] = data[coordinate_fields["latitude"]]
    data["longitude"] = data[coordinate_fields["longitude"]]
    data["elevation"] = data[coordinate_fields["elevation"]]

    data = data.set_index("utc_datetime")

    return ILVIS2DataFrame(data)



"""Utility routines."""

from __future__ import annotations

from pathlib import Path

from invoke import run

PROJECT_DIR = Path(__file__).parent.parent


def print_and_run(cmd, **run_kwargs):
    """Print the command string and then execute it."""
    print(cmd)
    kwargs = {
        "pty": True,
        **run_kwargs,
    }
    return run(cmd, **kwargs)



from __future__ import annotations

import datetime as dt
from typing import Literal

import pandera as pa
import pydantic
from earthaccess.results import DataGranule
from pandera.typing import DataFrame, Index, Series

from nsidc.iceflow.itrf import ITRF_REGEX


class CommonDataColumnsSchema(pa.DataFrameModel):
    utc_datetime: Index[pa.dtypes.DateTime] = pa.Field(check_name=True)
    ITRF: Series[str] = pa.Field(str_matches=ITRF_REGEX.pattern)
    latitude: Series[float] = pa.Field(coerce=True)
    longitude: Series[float] = pa.Field(coerce=True)
    elevation: Series[float] = pa.Field(coerce=True)
    # In practice, "dataset" is always available to provide provenance on where
    # each point comes from, but we mark it as optional here because it is not
    # strictly necessary for e.g., ITRF transformations.
    dataset: str | None


class ATM1BSchema(CommonDataColumnsSchema):
    # Data fields unique to ATM1B data.
    rel_time: Series[float] = pa.Field(nullable=True, coerce=True)
    xmt_sigstr: Series[float] = pa.Field(nullable=True, coerce=True)
    rcv_sigstr: Series[float] = pa.Field(nullable=True, coerce=True)
    azimuth: Series[float] = pa.Field(nullable=True, coerce=True)
    pitch: Series[float] = pa.Field(nullable=True, coerce=True)
    roll: Series[float] = pa.Field(nullable=True, coerce=True)
    gps_pdop: Series[float] = pa.Field(nullable=True, coerce=True)
    gps_time: Series[float] = pa.Field(nullable=True, coerce=True)
    passive_signal: Series[float] = pa.Field(nullable=True, coerce=True)
    passive_footprint_latitude: Series[float] = pa.Field(nullable=True, coerce=True)
    passive_footprint_longitude: Series[float] = pa.Field(nullable=True, coerce=True)
    passive_footprint_synthesized_elevation: Series[float] = pa.Field(
        nullable=True, coerce=True
    )
    pulse_width: Series[float] = pa.Field(nullable=True, coerce=True)


# Note: the ILVIS2 data contain multiple sets of lat/lon/elev. The common
# schema assumes one set of lat/lon/elev which is used for the ITRF
# transformation code.
class ILVIS2Schema(CommonDataColumnsSchema):
    """ILVIS2 Data Schema.

    Note that ILVIS2 data contain multiple sets of lat/lon/elev.

    * GLAT/GLON/GZ represent the center of the lowest mode in the waveform.
    * HLAT/HLON/HZ represent the center of the highest detected mode within the
      waveform. Both of these sets of lat/lon/elev are available across v1 and
      v2 ILVIS2 data.

    ILVIS V1 data:
    * CLAT/CLON/ZC represent the centroid of the corresponding LVIS Level-1B waveform.

    ILVIS V2 data:
    * TLAT/TLON/ZT, which represent the highest detected signal
    """

    # Common columns
    LFID: Series[float] = pa.Field(nullable=True, coerce=True)
    SHOTNUMBER: Series[float] = pa.Field(nullable=True, coerce=True)
    TIME: Series[float] = pa.Field(nullable=True, coerce=True)
    # ZG/GLAT/GLON: the center of the lowest detected mode within the waveform
    ZG: Series[float] = pa.Field(nullable=True, coerce=True)
    GLAT: Series[float] = pa.Field(nullable=True, coerce=True)
    GLON: Series[float] = pa.Field(nullable=True, coerce=True)
    # HLAT/HLON/ZH: the center of the highest detected mode within the waveform
    HLAT: Series[float] = pa.Field(nullable=True, coerce=True)
    HLON: Series[float] = pa.Field(nullable=True, coerce=True)
    ZH: Series[float] = pa.Field(nullable=True, coerce=True)

    # V104-specific
    # CLAT/CLON/ZC: Centroid of the corresponding LVIS Level-1B waveform
    CLAT: Series[float] = pa.Field(nullable=True, coerce=True)
    CLON: Series[float] = pa.Field(nullable=True, coerce=True)
    ZC: Series[float] = pa.Field(nullable=True, coerce=True)

    # V202B-specific
    AZIMUTH: Series[float] = pa.Field(nullable=True, coerce=True)
    CHANNEL_RH: Series[float] = pa.Field(nullable=True, coerce=True)
    CHANNEL_ZG: Series[float] = pa.Field(nullable=True, coerce=True)
    CHANNEL_ZT: Series[float] = pa.Field(nullable=True, coerce=True)
    COMPLEXITY: Series[float] = pa.Field(nullable=True, coerce=True)
    INCIDENT_ANGLE: Series[float] = pa.Field(nullable=True, coerce=True)
    RANGE: Series[float] = pa.Field(nullable=True, coerce=True)
    # RH%%%: Height (relative to ZG) at which % of the waveform energy occurs
    RH10: Series[float] = pa.Field(nullable=True, coerce=True)
    RH15: Series[float] = pa.Field(nullable=True, coerce=True)
    RH20: Series[float] = pa.Field(nullable=True, coerce=True)
    RH25: Series[float] = pa.Field(nullable=True, coerce=True)
    RH30: Series[float] = pa.Field(nullable=True, coerce=True)
    RH35: Series[float] = pa.Field(nullable=True, coerce=True)
    RH40: Series[float] = pa.Field(nullable=True, coerce=True)
    RH45: Series[float] = pa.Field(nullable=True, coerce=True)
    RH50: Series[float] = pa.Field(nullable=True, coerce=True)
    RH55: Series[float] = pa.Field(nullable=True, coerce=True)
    RH60: Series[float] = pa.Field(nullable=True, coerce=True)
    RH65: Series[float] = pa.Field(nullable=True, coerce=True)
    RH70: Series[float] = pa.Field(nullable=True, coerce=True)
    RH75: Series[float] = pa.Field(nullable=True, coerce=True)
    RH80: Series[float] = pa.Field(nullable=True, coerce=True)
    RH85: Series[float] = pa.Field(nullable=True, coerce=True)
    RH90: Series[float] = pa.Field(nullable=True, coerce=True)
    RH95: Series[float] = pa.Field(nullable=True, coerce=True)
    RH96: Series[float] = pa.Field(nullable=True, coerce=True)
    RH97: Series[float] = pa.Field(nullable=True, coerce=True)
    RH98: Series[float] = pa.Field(nullable=True, coerce=True)
    RH99: Series[float] = pa.Field(nullable=True, coerce=True)
    RH100: Series[float] = pa.Field(nullable=True, coerce=True)
    # Highest detected signal
    TLAT: Series[float] = pa.Field(nullable=True, coerce=True)
    TLON: Series[float] = pa.Field(nullable=True, coerce=True)
    ZT: Series[float] = pa.Field(nullable=True, coerce=True)

    class Config:
        # This ensures all columns are present, regardless of the date. Granules
        # before 2017 use the V104 fields and anything after uses the v202b
        # fields. The data type for all values must be `float` because the null
        # value is `np.nan` - a float.
        add_missing_columns = True


class GLAH06Schema(CommonDataColumnsSchema):
    # Note: all of these variables are extracted from the "Data_40HZ"
    # group. There is also a "Data_1HZ" group, which contains similar variables,
    # but does not include the elevation data and appears to be a downsampled
    # version of the "Data_40HZ" group. See
    # https://github.com/nsidc/nsidc-iceflow/issues/43.
    i_rec_ndx: Series[int] = pa.Field(coerce=True)
    i_shot_count: Series[int] = pa.Field(coerce=True)
    d_lat: Series[float]
    d_lon: Series[float]
    d_elev: Series[float]
    d_refRng: Series[float]
    d_dTrop: Series[float]
    d_satElevCorr: Series[float]
    d_GmC: Series[float]
    d_wTrop: Series[float]
    d_beamCoelv: Series[float]
    d_beamAzimuth: Series[float]
    d_SigBegOff: Series[float]
    d_TrshRngOff: Series[float]
    d_SigEndOff: Series[float]
    d_cntRngOff: Series[float]
    d_isRngOff: Series[float]
    d_siRngOff: Series[float]
    d_ldRngOff: Series[float]
    d_ocRngOff: Series[float]
    rng_uqf_sigbeg1_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_sigend1_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_thres1_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_cent1_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_sigbeg2_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_sigend2_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_thres2_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_cent2_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_is_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_si_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_ld_flg: Series[int] = pa.Field(coerce=True)
    rng_uqf_oc_flg: Series[int] = pa.Field(coerce=True)
    sat_corr_flg: Series[int] = pa.Field(coerce=True)
    elev_use_flg: Series[int] = pa.Field(coerce=True)
    att_pad_use_flg: Series[int] = pa.Field(coerce=True)
    att_calc_pad_flg: Series[int] = pa.Field(coerce=True)
    att_lpa_flg: Series[int] = pa.Field(coerce=True)
    sigma_att_flg: Series[int] = pa.Field(coerce=True)
    i_satNdx: Series[int] = pa.Field(coerce=True)
    d_pctSAT: Series[float]
    elv_cnt_1_flg: Series[int] = pa.Field(coerce=True)
    elv_cnt_2_flg: Series[int] = pa.Field(coerce=True)
    elv_peak_1_flg: Series[int] = pa.Field(coerce=True)
    elv_peak_2_flg: Series[int] = pa.Field(coerce=True)
    elv_thres_flg: Series[int] = pa.Field(coerce=True)
    elv_gauss_flg: Series[int] = pa.Field(coerce=True)
    elv_other_flg: Series[int] = pa.Field(coerce=True)
    elv_cloud_flg: Series[int] = pa.Field(coerce=True)
    d_TxNrg: Series[float]
    d_d2refTrk: Series[float]
    d_DEM_elv: Series[float]
    d_ocElv: Series[float]
    d_poTide: Series[float]
    d_gdHt: Series[float]
    d_erElv: Series[float]
    d_eqElv: Series[float]
    d_ldElv: Series[float]
    d_deltaEllip: Series[float]
    d_ElevBiasCorr: Series[float]
    i_DEM_hires_src_1: Series[int] = pa.Field(coerce=True)
    d_reflctUC: Series[float]
    d_sDevNsOb1: Series[float]
    d_satNrgCorr: Series[float]
    d_RecNrgAll: Series[float]
    d_skew2: Series[float]
    d_kurt2: Series[float]
    d_maxRecAmp: Series[float]
    d_maxSmAmp: Series[float]
    i_nPeaks1: Series[int] = pa.Field(coerce=True)
    i_numPk: Series[int] = pa.Field(coerce=True)
    i_gval_rcv: Series[int] = pa.Field(coerce=True)
    d_FRir_cldtop: Series[float]
    FRir_qa_flg: Series[int] = pa.Field(coerce=True)
    d_FRir_intsig: Series[float]


IceflowDataFrame = DataFrame[CommonDataColumnsSchema]
ATM1BDataFrame = DataFrame[ATM1BSchema]
ILVIS2DataFrame = DataFrame[ILVIS2Schema]
GLAH06DataFrame = DataFrame[GLAH06Schema]

ATM1BShortName = Literal["ILATM1B", "BLATM1B"]
DatasetShortName = ATM1BShortName | Literal["ILVIS2"] | Literal["GLAH06"]


class Dataset(pydantic.BaseModel):
    short_name: DatasetShortName
    version: str

    @property
    def subdir_name(self):
        return f"{self.short_name}_{self.version}"


class ATM1BDataset(Dataset):
    short_name: ATM1BShortName


# This mirrors the bounding box construct in `earthaccess` and `icepyx`: a
# list/float of len 4:
# (lower_left_lon, lower_left_lat, upper_right_lon, upper_right_lat)
BoundingBoxLike = list[float] | tuple[float, float, float, float]


TemporalRange = tuple[dt.datetime | dt.date, dt.datetime | dt.date]


class IceflowSearchResult(pydantic.BaseModel):
    dataset: Dataset
    granules: list[DataGranule]

    # Pydantic can't infer what `DataGranule` is. This ignores validation for
    # this that type
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    @property
    def total_size_mb(self):
        granule_sizes_mb = [granule.size() for granule in self.granules]
        return sum(granule_sizes_mb)


IceflowSearchResults = list[IceflowSearchResult]



from __future__ import annotations

from nsidc.iceflow.data.models import (
    Dataset,
)
from nsidc.iceflow.data.supported_datasets import (
    ALL_SUPPORTED_DATASETS,
    BLATM1BDataset,
    GLAH06Dataset,
    ILATM1BDataset,
    ILVIS2Dataset,
)

__all__ = [
    "ALL_SUPPORTED_DATASETS",
    "BLATM1BDataset",
    "Dataset",
    "GLAH06Dataset",
    "ILATM1BDataset",
    "ILVIS2Dataset",
]



from __future__ import annotations

import gps_timemachine


def test_gps_timemachine_import():
    assert gps_timemachine is not None


def test_import_package():
    from nsidc import iceflow

    assert iceflow is not None



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from pyproj import Transformer

from nsidc.iceflow.data.models import IceflowDataFrame
from nsidc.iceflow.itrf.converter import (
    _datetime_to_decimal_year,
    _itrf_transformation_step,
    transform_itrf,
)


def test_transform_itrf():
    synth_df = pd.DataFrame(
        {
            # Note: duplicate data here because otherwise a deprecation warning
            # error about single-value series is raised from pandas' internals,
            # and pytest complains.
            "latitude": [70, 70],
            "longitude": [-50, -50],
            "elevation": [1, 1],
            "ITRF": ["ITRF93", "ITRF93"],
            "dataset": ["foov1", "foov2"],
        },
    )

    constant_datetime = pd.to_datetime("1993-07-02 12:00:00")
    synth_df.index = pd.Index([constant_datetime] * 2, name="utc_datetime")
    synth_iceflow_df = IceflowDataFrame(synth_df)

    result = transform_itrf(
        data=synth_iceflow_df,
        target_itrf="ITRF2014",
    )

    # Before and after values were extracted from the synthetic dataset in the
    # ITRF corrections notebook put together by Kevin. See:
    # ./docs/notebooks/corrections.ipynb
    assert result.latitude.to_numpy()[0] == 69.99999974953995
    assert result.longitude.to_numpy()[0] == -50.0000001319163
    # Rounding to 8 decimals in a field assumed to be measured in meters is 10nm accuracy!
    # This rounding became necessary when pyproj updated to a newer than 3.6.1
    # version, leading to a difference of ~1nm in the resulting calculation.
    assert np.round(result.elevation.to_numpy()[0], 8) == np.round(
        1.0052761882543564, 8
    )


def _itrf_cart_coords_to_wgs84(x, y, z):
    transformer = Transformer.from_pipeline("+proj=cart +inv +ellps=WGS84")

    lon, lat, elev = transformer.transform(x, y, z)

    return lon, lat, elev


def test_transform_itrf_ectt():
    """Verifies that our transformation code matches the example transformation given by the ECTT.

    The ETRF/ITRF Coordinate Transformation Tool (ECTT)
    (https://doi.org/10.24414/ROB-EUREF-ECTT) provides a web-based interface for
    converting between various ITRF realizations at different epochs.

    See the ECTT here: https://epncb.oma.be/_productsservices/coord_trans/index.php#results

    For this test, use the existing "station_1" example given in the ECTT. Set
    input frame to ITRF93 and Epoch to 1993.0 and set "Transform to" to ITRF2014
    with an Epoch of 1993.0.

    Input:
    # Lines starting by # are treated as comments
    # Fields (in decimal format) should be separated by at least one space
    #
    # --> Example without velocity <--
    # Stationname (no space character) X[m] Y[m] Z[m] :
    Station_1 4027894.006 307045.600 4919474.910

    Results:
    #_Station Frame  Epoch        X[m]          Y[m]          Z[m]    VX[m/yr] VY[m/yr] VZ[m/yr]
    Station_1 ITRF2014 1993.00  4027894.0021   307045.5873  4919474.9151
    """
    # Convert Station 1 initial coordinates, given in ITRF 1993 meters, to
    # WGS84 lat/lon/elev for input into the iceflow conversion function.
    lon, lat, elev = _itrf_cart_coords_to_wgs84(4027894.006, 307045.600, 4919474.910)

    synth_df = pd.DataFrame(
        {
            # Note: duplicate data here because otherwise a deprecation warning
            # error about single-value series is raised from pandas' internals,
            # and pytest complains.
            "longitude": [lon, lon],
            "latitude": [lat, lat],
            "elevation": [elev, elev],
            "ITRF": ["ITRF93", "ITRF93"],
            "dataset": ["foov1", "foov2"],
        },
    )

    constant_datetime = pd.to_datetime("1993-01-01 00:00:00")
    synth_df.index = pd.Index([constant_datetime] * 2, name="utc_datetime")
    synth_iceflow_df = IceflowDataFrame(synth_df)

    result = transform_itrf(
        data=synth_iceflow_df,
        target_itrf="ITRF2014",
    )

    # convert station 1 results in ITRF 2014 meters to WGS84 lat/lon/elev
    expected_lon, expected_lat, expected_elev = _itrf_cart_coords_to_wgs84(
        4027894.0021, 307045.5873, 4919474.9151
    )

    # The results are very close to what we expect with the full pipeline
    assert np.round(result.longitude.to_numpy()[0], 9) == np.round(expected_lon, 9)
    assert np.round(result.latitude.to_numpy()[0], 8) == np.round(expected_lat, 8)
    assert np.round(result.elevation.to_numpy()[0], 4) == np.round(expected_elev, 4)


def test__itrf_transformation_step():
    """Tests that we have the ITRF transformation correct.


    The ETRF/ITRF Coordinate Transformation Tool (ECTT)
    (https://doi.org/10.24414/ROB-EUREF-ECTT) provides a web-based interface for
    converting between various ITRF realizations at different epochs.

    See the ECTT here: https://epncb.oma.be/_productsservices/coord_trans/index.php#results

    For this test, use the existing "station_1" example given in the ECTT. Set
    input frame to ITRF93 and Epoch to 1993.0 and set "Transform to" to ITRF2014
    with an Epoch of 1993.0.

    Input:
    # Lines starting by # are treated as comments
    # Fields (in decimal format) should be separated by at least one space
    #
    # --> Example without velocity <--
    # Stationname (no space character) X[m] Y[m] Z[m] :
    Station_1 4027894.006 307045.600 4919474.910

    Results:
    #_Station Frame  Epoch        X[m]          Y[m]          Z[m]    VX[m/yr] VY[m/yr] VZ[m/yr]
    Station_1 ITRF2014 1993.00  4027894.0021   307045.5873  4919474.9151
    """
    # Input values given to ECTT
    station_1_x = 4027894.006
    station_1_y = 307045.600
    station_1_z = 4919474.910
    station_1_epoch = 1993.0

    transformation_step = _itrf_transformation_step("ITRF93", "ITRF2014")
    transformer = Transformer.from_pipeline("+proj=pipeline " + transformation_step)

    actual_x, actual_y, actual_z, _ = transformer.transform(
        station_1_x, station_1_y, station_1_z, station_1_epoch
    )

    # Exact output values reported by ECTT
    expected_x = 4027894.0021
    expected_y = 307045.5873
    expected_z = 4919474.9151

    assert np.round(actual_x, 4) == expected_x
    assert np.round(actual_y, 4) == expected_y
    assert np.round(actual_z, 4) == expected_z


def test__itrf_transformation_step_failure():
    """Test that the _itrf_transformation_step raises an error if there is no
    pre-defined transform."""

    with pytest.raises(
        RuntimeError, match="Failed to find a pre-defined ITRF transformation"
    ):
        _itrf_transformation_step("ITRF2022", "ITRF93")


def test__itrf_transformation_step_fwd():
    """Test that the forward transform is returned."""
    expected = "+step +init=ITRF2014:ITRF2008"
    actual_step = _itrf_transformation_step("ITRF2014", "ITRF2008")

    assert expected == actual_step


def test__itrf_transformation_step_inv():
    """Test that the forward transform is returned."""
    expected = "+step +inv +init=ITRF2014:ITRF2008"
    actual_step = _itrf_transformation_step("ITRF2008", "ITRF2014")

    assert expected == actual_step


@pytest.mark.parametrize("timezone", ["America/Denver", "UTC"])
def test__datetime_to_decimal_year(timezone, monkeypatch):
    monkeypatch.setenv("TZ", timezone)
    result = _datetime_to_decimal_year(pd.to_datetime("1993-07-02 12:00:00"))
    assert result == 1993.5



from __future__ import annotations

import datetime as dt
import logging
import re
from enum import Enum
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
import pandera as pa
from gps_timemachine.gps import leap_seconds
from numpy.typing import DTypeLike

from nsidc.iceflow.data.models import ATM1BDataFrame

"""
The dtypes used to read any of the input ATM1B input files.

The ATM1B QFIT format:
https://nsidc.org/sites/nsidc.org/files/files/ReadMe_qfit.txt
"""
ATM1B_DTYPE_10_BE = np.dtype(
    [
        ("rel_time", ">i4"),
        ("latitude", ">i4"),
        ("longitude", ">i4"),
        ("elevation", ">i4"),
        ("xmt_sigstr", ">i4"),
        ("rcv_sigstr", ">i4"),
        ("azimuth", ">i4"),
        ("pitch", ">i4"),
        ("roll", ">i4"),
        ("gps_time", ">i4"),
    ]
)
ATM1B_DTYPE_10_LE = ATM1B_DTYPE_10_BE.newbyteorder("<")

ATM1B_DTYPE_12_BE = np.dtype(
    [
        ("rel_time", ">i4"),
        ("latitude", ">i4"),
        ("longitude", ">i4"),
        ("elevation", ">i4"),
        ("xmt_sigstr", ">i4"),
        ("rcv_sigstr", ">i4"),
        ("azimuth", ">i4"),
        ("pitch", ">i4"),
        ("roll", ">i4"),
        ("gps_pdop", ">i4"),
        ("pulse_width", ">i4"),
        ("gps_time", ">i4"),
    ]
)
ATM1B_DTYPE_12_LE = ATM1B_DTYPE_12_BE.newbyteorder("<")

ATM1B_DTYPE_14_BE = np.dtype(
    [
        ("rel_time", ">i4"),
        ("latitude", ">i4"),
        ("longitude", ">i4"),
        ("elevation", ">i4"),
        ("xmt_sigstr", ">i4"),
        ("rcv_sigstr", ">i4"),
        ("azimuth", ">i4"),
        ("pitch", ">i4"),
        ("roll", ">i4"),
        ("passive_signal", ">i4"),
        ("passive_footprint_latitude", ">i4"),
        ("passive_footprint_longitude", ">i4"),
        ("passive_footprint_synthesized_elevation", ">i4"),
        ("gps_time", ">i4"),
    ]
)
ATM1B_DTYPE_14_LE = ATM1B_DTYPE_14_BE.newbyteorder("<")


class Endian(Enum):
    LITTLE = 1
    BIG = 2


def _data_dtype(endianness: Endian, field_count: int) -> DTypeLike:
    """Return the appropriate QFIT dtype based on the given endianness and
    number of fields in the file."""
    return {
        Endian.LITTLE: {
            10: ATM1B_DTYPE_10_LE,
            12: ATM1B_DTYPE_12_LE,
            14: ATM1B_DTYPE_14_LE,
        },
        Endian.BIG: {
            10: ATM1B_DTYPE_10_BE,
            12: ATM1B_DTYPE_12_BE,
            14: ATM1B_DTYPE_14_BE,
        },
    }[endianness][field_count]


def _file_dtype(filepath: Path) -> DTypeLike:
    """Return the dtype for the given file."""
    data_endianness = Endian.BIG

    record_size = np.fromfile(filepath, dtype=">i4", count=1)[0]
    if record_size >= 100:
        record_size = np.fromfile(filepath, dtype="<i4", count=1)[0]
        if record_size >= 100:
            raise ValueError("invalid record size found")
        data_endianness = Endian.LITTLE

    field_count = int(record_size / 4)

    dtype = _data_dtype(data_endianness, field_count)

    return dtype


def _blatm1bv1_date(fn) -> dt.date:
    """Return the date from the given BLATM1B filename."""
    fn_date = None

    m = re.search(r"BLATM1B_(\d{8})", fn)
    if m:
        fn_date_str = m.group(1)
    else:
        m = re.search(r"BLATM1B_(\d{6})", fn)
        if not m:
            err_msg = f"Failed to extract date from BLATM1B v1 file: {fn}"
            raise RuntimeError(err_msg)

        fn_date_str = m.group(1)
        if int(fn_date_str[:2]) > 9:
            fn_date_str = "19" + fn_date_str
        else:
            fn_date_str = "20" + fn_date_str

    fn_date = dt.datetime.strptime(fn_date_str, "%Y%m%d").date()

    return fn_date


def _ilatm1b_date(fn: str) -> dt.date:
    """Return the date from the given ILATM1B filename."""
    m = re.search(r"_(\d{8})_", fn)
    if not m:
        err = f"Failed to extract date from filepath: {fn}"
        raise RuntimeError(err)
    fn_date = m.group(1)
    return dt.datetime.strptime(fn_date, "%Y%m%d").date()


def _shift_lon(lon):
    """Shifts longitude values from [0,360] to [-180,180]"""
    if lon >= 180.0:
        return lon - 360.0
    return lon


def _augment_with_optional_values(df, original_shape):
    """Add columns (w/ `np.nan`) to the dataframe depending on what fields
    the original data did not include.
    """
    rows, cols = original_shape
    missing = np.full(shape=(rows,), fill_value=np.nan)

    if cols == 10:
        df["gps_pdop"] = missing
        df["pulse_width"] = missing
        df["passive_signal"] = missing
        df["passive_footprint_latitude"] = missing
        df["passive_footprint_longitude"] = missing
        df["passive_footprint_synthesized_elevation"] = missing
    elif cols == 12:
        df["passive_signal"] = missing
        df["passive_footprint_latitude"] = missing
        df["passive_footprint_longitude"] = missing
        df["passive_footprint_synthesized_elevation"] = missing
    elif cols == 14:
        df["gps_pdop"] = missing
        df["pulse_width"] = missing
    else:
        raise ValueError("Unknown number of columns: cannot augment")


def _strip_header(data):
    """Slice the header from the given data; skip the first row because we
    know it's a valid header; any rows with negative first elements
    are also header rows.
    """
    idx = 1
    while data[idx][0] < 0:
        idx += 1

    return data[idx:]


def _utc_datetime(
    gps_time: pd.Series[int], file_date: dt.date
) -> pd.Series[pd.Timestamp]:
    """Return `utc_datetime` Series, with values calculated from the given
    date and the GPS time values, with a leap second adjustment to the GPS
    times.
    """
    tdf = pd.DataFrame()
    tdf["hour"] = (gps_time / 1e7).astype(np.uint8)
    tdf["minute"] = ((gps_time % 1e7) / 1e5).astype(np.uint8)
    tdf["second"] = ((gps_time % 1e5) / 1e3).astype(np.uint8)
    tdf["millisecond"] = (gps_time % 1e3).astype(np.uint16)
    tdf["year"] = file_date.year
    tdf["month"] = file_date.month
    tdf["day"] = file_date.day

    ls = int(leap_seconds(dt.datetime(file_date.year, file_date.month, file_date.day)))
    utc = pd.to_datetime(tdf).to_numpy() - np.timedelta64(ls, "s")

    return pd.Series(utc)


def _atm1b_qfit_dataframe(filepath: Path) -> pd.DataFrame:
    """Read an ATM1B QFIT file into a DataFrame, stripping bad data if
    necessary.
    """
    dtype = _file_dtype(filepath)

    raw_data = np.fromfile(filepath, dtype=dtype)
    raw_data = _strip_header(raw_data)

    if dtype in (ATM1B_DTYPE_14_LE, ATM1B_DTYPE_14_BE):
        # Ignore records with invalid data (this occurs in the 14-word
        # format records containing passive brightness data)
        logging.info("Before filter. Data shape: %s", raw_data.shape)
        raw_data = raw_data[
            (raw_data["latitude"] != 0) & (raw_data["elevation"] != -9999)
        ]
        logging.info("After filter. Data shape: %s", raw_data.shape)

        if raw_data.shape[0] == 0:
            logging.warning("After removal of bad data, file contains no valid data.")
            return pd.DataFrame()

    return pd.DataFrame(raw_data)


def _atm1b_qfit_data(filepath: Path, file_date: dt.date) -> pd.DataFrame:
    """Returns an ATM1B DataFrame read from a QFIT file, performing all
    necessary conversions on the data.
    """
    df = _atm1b_qfit_dataframe(filepath)
    original_shape = df.shape

    df["latitude"] = df["latitude"] * 1e-6
    df["longitude"] = df["longitude"] * 1e-6
    df["longitude"] = df["longitude"].apply(_shift_lon)
    # elevation values are natively stored as Meters 10**-3. We convert them
    # back to meters here.
    df["elevation"] = df["elevation"] * 1e-3
    df["elevation"] = df["elevation"].astype(np.float32)
    df["utc_datetime"] = _utc_datetime(df["gps_time"], file_date)
    _augment_with_optional_values(df, original_shape)

    return df


# TODO: extract this for reuse with other data products.
def _normalize_itrf_str(itrf_str: str) -> str:
    """Normalizes common ITRF strings into ones recognizable by proj.

    E.g., "ITRF2020" is common, but proj only recognizes "ITRF20".
    """
    itrf_str = itrf_str.upper()
    try:
        itrf_str = {
            "ITRF00": "ITRF2000",
            "ITRF05": "ITRF2005",
            "ITRF08": "ITRF2008",
            "ITRF2020": "ITRF20",
        }[itrf_str]
    except KeyError:
        pass

    return itrf_str


def _qfit_file_header(filepath: Path) -> str:
    """Return the header string from a QFIT file."""
    record_size = np.fromfile(filepath, dtype=">i4", count=1)[0]
    # The header length for each record is the number of bytes after the first
    # word.
    header_size = record_size - 4
    dtype = np.dtype([("record_type", ">i4"), ("header", f">S{header_size}")])

    if record_size >= 100:
        record_size = np.fromfile(filepath, dtype="<i4", count=1)[0]
        if record_size >= 100:
            raise ValueError("invalid record size found")
        header_size = record_size - 4
        dtype = np.dtype([("record_type", "<i4"), ("header", f"<S{header_size}")])

    raw_data = np.fromfile(filepath, dtype=dtype)

    # In 'normal' files with headers, we skip the first two records
    # and only keep those whose record_type is negative.
    if len(raw_data) > 2:
        idx = 2
        while raw_data[idx]["record_type"] < 0:
            idx += 1
        header = raw_data[2:idx]
        return "".join([r["header"].decode("UTF-8") for r in header])

    err = "Failed to read qfit file header for {filepath}"
    raise RuntimeError(err)


def _infer_qfit_itrf(filepath: Path) -> str:
    """Takes an ILATM1B/BLATM1B qfit filepath and returns a string representing
    the ITRF.

    This function infers the ITRF based on the qfit file header, which is
    described here:
    https://nsidc.org/sites/nsidc.org/files/files/ReadMe_qfit.txt

    The string we extract the ITRF from looks like this:

        `./091109_aa_l12_cfm_itrf05_18may10_palm_roth_amu2`

    From which we extract an ITRF of `ITRF2005` from the `_itrf05_` bit.

    According to Michael Studinger (see
    https://github.com/nsidc/nsidc-iceflow/issues/35#issuecomment-2408619586), This
    string represents the "GPS trajectory that was used to reference the lidar
    data and that has the ITRF epoch in its file name."
    """
    header = _qfit_file_header(filepath)
    results = re.finditer(r"itrf\d{2,4}", header)
    itrfs = list({result.group() for result in results})

    if len(itrfs) == 1:
        itrf_from_qfit_header = _normalize_itrf_str(itrfs[0])
        return itrf_from_qfit_header

    err_msg = f"Failed to extract ITRF from qfit header: {filepath}"
    raise RuntimeError(err_msg)


def _extract_itrf_from_h5_file(filepath: Path) -> str:
    """Extract ITRF from the h5 filepath (ILATM1B v2)."""
    with h5py.File(filepath, "r") as ds:
        itrf = str(ds["ancillary_data"]["reference_frame"][:][0], encoding="utf8")
    itrf = _normalize_itrf_str(itrf)

    return itrf


def extract_itrf(filepath: Path) -> str:
    ext = filepath.suffix

    if ext == ".qi":
        return _infer_qfit_itrf(filepath)
    elif ext == ".h5":
        return _extract_itrf_from_h5_file(filepath)

    err = f"Failed to read ITRF from unrecognized file: {filepath}"
    raise RuntimeError(err)


def _ilatm1bv2_dataframe(filepath: Path) -> pd.DataFrame:
    """Returns an ATM1B DataFrame read from an HDF5 file, performing all
    necessary scaling and type conversion.
    """
    variables = [
        ("rel_time", "instrument_parameters/rel_time", 1000, np.int32),
        ("latitude", "latitude", None, None),
        ("longitude", "longitude", None, None),
        ("elevation", "elevation", None, None),
        ("xmt_sigstr", "instrument_parameters/xmt_sigstr", None, None),
        ("rcv_sigstr", "instrument_parameters/rcv_sigstr", None, None),
        ("azimuth", "instrument_parameters/azimuth", 1000, np.int32),
        ("pitch", "instrument_parameters/pitch", 1000, np.int32),
        ("roll", "instrument_parameters/roll", 1000, np.int32),
        ("gps_pdop", "instrument_parameters/gps_pdop", 10, np.int32),
        ("pulse_width", "instrument_parameters/pulse_width", 1, np.uint32),
        ("gps_time", "instrument_parameters/time_hhmmss", 1000, np.uint32),
    ]
    df = pd.DataFrame()
    with h5py.File(filepath, "r") as atmv2:
        for key, name, scale_factor, dtype in variables:
            if scale_factor and dtype:
                df[key] = (atmv2[name][:] * scale_factor).astype(dtype)
            else:
                df[key] = atmv2[name][:]

    return df


def _ilatm1bv2_data(fn: Path, file_date: dt.date) -> pd.DataFrame:
    """Returns an ATM1B DataFrame, performing all necessary conversions /
    augmentation on the data.
    """
    df = _ilatm1bv2_dataframe(fn)
    original_shape = df.shape

    df["longitude"] = df["longitude"].apply(_shift_lon)
    df["utc_datetime"] = _utc_datetime(df["gps_time"], file_date)
    _augment_with_optional_values(df, original_shape)

    return df


@pa.check_types()
def atm1b_data(filepath: Path) -> ATM1BDataFrame:
    """
    Return the atm1b data given a filename.

    Parameters
    ----------
    filepath
        The filepath to read.

    Returns
    -------
    data
        The atm1b (pandas.DataFrame) data.
    """
    # Example filenames:
    # ILATM1B_20140430_110310.ATM4BT4.h5
    # ILATM1B_20111104_181304.ATM4BT4.qi
    # BLATM1B_20060522_145449.qi
    # BLATM1B_20041127atm2_210316jr.lutF.qi
    filename = filepath.name

    # Find the date, which corresponds to the product version.
    match = re.search(r".*_(\d{4})\d{4}.*", filename)
    if not match:
        err = f"Failed to recognize {filename} as ATM1B data."
        raise RuntimeError(err)

    year = int(match.group(1))

    if year >= 2013:
        file_date = _ilatm1b_date(filename)
        data = _ilatm1bv2_data(filepath, file_date)
        dataset = "ILATM1Bv2"
    elif year >= 2009:
        file_date = _ilatm1b_date(filename)
        data = _atm1b_qfit_data(filepath, file_date)
        dataset = "ILATM1Bv1"
    else:
        file_date = _blatm1bv1_date(filename)
        data = _atm1b_qfit_data(filepath, file_date)
        dataset = "BLATM1Bv1"

    itrf = extract_itrf(filepath)
    data["ITRF"] = itrf
    data["dataset"] = dataset

    data = data.set_index("utc_datetime")

    return ATM1BDataFrame(data)



from __future__ import annotations

from pathlib import Path

import earthaccess
import pydantic
from loguru import logger

from nsidc.iceflow.data.models import (
    Dataset,
    IceflowSearchResult,
    IceflowSearchResults,
)
from nsidc.iceflow.data.supported_datasets import ALL_SUPPORTED_DATASETS


def _find_iceflow_data_for_dataset(
    *,
    dataset: Dataset,
    **search_kwargs,
) -> IceflowSearchResult:
    earthaccess.login()

    ctx_string = f"{dataset.short_name=} {dataset.version=} with {search_kwargs=}"

    try:
        granules_list = earthaccess.search_data(
            short_name=dataset.short_name,
            version=dataset.version,
            # Select only cloud-hosted granules. Without specifying cloud vs
            # non-cloud, we may get duplicate granules as long as the ECS copy
            # remains.
            cloud_hosted=True,
            **search_kwargs,
        )
    except IndexError:
        # There's no data matching the given parameters.
        granules_list = []

    num_results = len(granules_list)

    if not num_results:
        logger.warning(f"Found no results for {ctx_string}")
        granules_list = []

    iceflow_search_result = IceflowSearchResult(dataset=dataset, granules=granules_list)
    return iceflow_search_result


def _download_iceflow_search_result(
    *,
    iceflow_search_result: IceflowSearchResult,
    output_dir: Path,
) -> list[Path]:
    # No granules found for this search result object.
    if not iceflow_search_result.granules:
        return []
    # short_name and version-based subdir for data.
    subdir_name = f"{iceflow_search_result.dataset.short_name}_{iceflow_search_result.dataset.version}"
    output_subdir = output_dir / subdir_name

    logger.info(
        f"Downloading {len(iceflow_search_result.granules)} granules"
        f" (approx. {iceflow_search_result.total_size_mb} MB)"
        f" to {output_subdir}."
    )

    output_subdir.mkdir(exist_ok=True)
    downloaded_files = earthaccess.download(
        iceflow_search_result.granules, str(output_subdir)
    )
    downloaded_filepaths = [Path(filepath_str) for filepath_str in downloaded_files]
    # There may be duplicate filepaths returned by earthaccess because of data
    # existing both in the cloud and on ECS.
    downloaded_filepaths = list(set(downloaded_filepaths))

    return downloaded_filepaths


@pydantic.validate_call()
def find_iceflow_data(
    *,
    datasets: list[Dataset] = ALL_SUPPORTED_DATASETS,
    **search_kwargs,
) -> IceflowSearchResults:
    """Find iceflow-compatible data using search kwargs.

    `search_kwargs` are passed to `earthaccess.search_data`, allowing for
    CMR-supported filters (see
    https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html)
    """
    iceflow_search_results = []
    for dataset in datasets:
        iceflow_search_result = _find_iceflow_data_for_dataset(
            dataset=dataset,
            **search_kwargs,
        )
        iceflow_search_results.append(iceflow_search_result)

    return iceflow_search_results


def download_iceflow_results(
    iceflow_search_results: IceflowSearchResults,
    output_dir: Path,
) -> list[Path]:
    all_downloaded_files = []
    for iceflow_search_result in iceflow_search_results:
        downloaded_filepaths = _download_iceflow_search_result(
            iceflow_search_result=iceflow_search_result,
            output_dir=output_dir,
        )
        all_downloaded_files.extend(downloaded_filepaths)

    return all_downloaded_files



"""Module with functions defining the user-facing API for nsidc-iceflow."""

from __future__ import annotations

import shutil
from pathlib import Path

import dask.dataframe as dd
from loguru import logger

from nsidc.iceflow.data.ilvis2 import ILVIS2_DEFAULT_COORDINATE_SET
from nsidc.iceflow.data.read import read_iceflow_datafiles
from nsidc.iceflow.data.supported_datasets import ALL_SUPPORTED_DATASETS
from nsidc.iceflow.itrf.converter import transform_itrf


def make_iceflow_parquet(
    *,
    data_dir: Path,
    target_itrf: str,
    overwrite: bool = False,
    target_epoch: str | None = None,
    ilvis2_coordinate_set=ILVIS2_DEFAULT_COORDINATE_SET,
) -> Path:
    """Create a parquet dataset containing the lat/lon/elev data in `data_dir`.

    This function creates a parquet dataset that can be easily used alongside
    dask, containing lat/lon/elev data. Users who are interested in the full
    data record with all fields provided by data in the `data_dir` should use
    `read_iceflow_datafiles`.

    Note: this function writes a single `iceflow.parquet` to the output
    dir. This code does not currently support updates to the parquet after being
    written. This is intended to help facilitate analysis of a specific area
    over time. If an existing `iceflow.parquet` exists and the user wants to
    create a new `iceflow.parquet` for a different area or timespan, they will
    need to move/remove the existing `iceflow.parquet` first (e.g., with the
    `overwrite=True` kwarg).
    """
    parquet_subdir = data_dir / "iceflow.parquet"
    if parquet_subdir.exists():
        if overwrite:
            logger.info("Removing existing iceflow.parquet")
            shutil.rmtree(parquet_subdir)
        else:
            raise RuntimeError(
                "An iceflow parquet file already exists. Use `overwrite=True` to overwrite."
            )

    all_subdirs = [
        data_dir / ds.subdir_name
        for ds in ALL_SUPPORTED_DATASETS
        if (data_dir / ds.subdir_name).is_dir()
    ]
    for subdir in all_subdirs:
        iceflow_filepaths = [path for path in subdir.iterdir() if path.is_file()]
        iceflow_df = read_iceflow_datafiles(
            iceflow_filepaths,
            ilvis2_coordinate_set=ilvis2_coordinate_set,
        )

        iceflow_df = transform_itrf(
            data=iceflow_df,
            target_itrf=target_itrf,
            target_epoch=target_epoch,
        )

        # Add a string col w/ dataset name and version.
        common_columns = ["latitude", "longitude", "elevation", "dataset"]
        common_dask_df = dd.from_pandas(iceflow_df[common_columns])
        if parquet_subdir.exists():
            dd.to_parquet(
                df=common_dask_df,
                path=parquet_subdir,
                append=True,
                ignore_divisions=True,
            )
        else:
            dd.to_parquet(
                df=common_dask_df,
                path=parquet_subdir,
            )

    return parquet_subdir



from __future__ import annotations

import calendar
import datetime as dt
from typing import cast

import pandas as pd
import pandera as pa
import pyproj
from shapely.geometry.point import Point

from nsidc.iceflow.data.models import IceflowDataFrame
from nsidc.iceflow.itrf import check_itrf
from nsidc.iceflow.itrf.plate_boundaries import plate_name


def _datetime_to_decimal_year(date):
    """Stolen from
    https://stackoverflow.com/questions/6451655/python-how-to-convert-datetime-dates-to-decimal-years,
    with one modification: `calendar.timegm` is used to set the epoch instead of
    `time.mktime`, which assumes local time.
    """

    def sinceEpoch(date):
        # returns seconds since epoch
        return calendar.timegm(date.timetuple())

    s = sinceEpoch

    year = date.year
    startOfThisYear = dt.datetime(year=year, month=1, day=1)
    startOfNextYear = dt.datetime(year=year + 1, month=1, day=1)

    yearElapsed = s(date) - s(startOfThisYear)
    yearDuration = s(startOfNextYear) - s(startOfThisYear)
    fraction = yearElapsed / yearDuration

    return date.year + fraction


def _check_valid_proj_step(proj_str) -> bool:
    """Check if the source/target ITRF pair can be expanded.

    Returns `True` if the combination is valid. Otherwise `False`.

    The combination is valid only if there is an init file on the proj data path
    matching the `source_itrf` that has an entry matching the `target_itrf`. See
    https://proj.org/en/9.3/resource_files.html#init-files for more info.
    """
    try:
        pyproj.Transformer.from_pipeline(proj_str)
        return True
    except pyproj.exceptions.ProjError:
        return False


def _itrf_transformation_step(source_itrf: str, target_itrf: str) -> str:
    """Get the ITRF transformation step for the given source/target ITRF.

    The transformation step returned by this function performs a helmert
    transform (see
    https://proj.org/en/9.4/operations/transformations/helmert.html).

    The parameters for the helmert transform come from proj init files (see
    https://proj.org/en/9.3/resource_files.html#init-files). For example,
    `+init=ITRF2014:ITRF2008` looks up the ITRF2008 helmert transformation step
    in the ITRF2014 data file (see
    https://github.com/OSGeo/PROJ/blob/master/data/ITRF2014).
    """
    # The `+inv` reverses the transformation. So `+init=ITRF2014:ITRF2008`
    # performs a helmert transform from ITRF2008 to ITRF2014.  This is the most
    # common case for `nsidc-iceflow`, because we tend to be targeting pre-icesat2
    # data for transformation to ITRF2014 (icesat2), so try this first.
    inv_itrf_transformation_step = f"+step +inv +init={target_itrf}:{source_itrf}"
    if _check_valid_proj_step(inv_itrf_transformation_step):
        return inv_itrf_transformation_step

    # Forward helmert transformation. `+init=ITRF2014:ITRF2008`
    # performs a helmert transform from ITRF2014 to ITRF2008.
    fwd_itrf_transformation_step = f"+step +init={source_itrf}:{target_itrf}"
    if _check_valid_proj_step(fwd_itrf_transformation_step):
        return fwd_itrf_transformation_step

    # There may not be a pre-defined helmert transformation. The user may want
    # to craft their own transformation pipeline.
    err_msg = (
        f"Failed to find a pre-defined ITRF transformation between {source_itrf} and {target_itrf}."
        " ITRF transformation parameters are provided by proj's ITRF init files."
        " Consider upgrading proj to ensure the latest data is available and try again."
        " See https://proj.org/en/latest/resource_files.html#init-files for more information."
        f" If no pre-defined transformation is available for {source_itrf} -> {target_itrf},"
        " it may be possible to define your own transformation using parameters found at https://itrf.ign.fr/."
        " See https://proj.org/en/latest/operations/transformations/helmert.html for more information."
    )
    raise RuntimeError(err_msg)


@pa.check_types()
def transform_itrf(
    data: IceflowDataFrame,
    target_itrf: str,
    target_epoch: str | None = None,
    # If a target epoch is given, the plate name can be given. If a target_epoch
    # is given but the plate name is not, each source ITRF is grouped together
    # and the mean of that chunk is used to determine the plate name.
    plate: str | None = None,
) -> IceflowDataFrame:
    """Transform the data's latitude/longitude/elevation variables from the
    source ITRF to the target ITRF.

    If a `target_epoch` is given, coordinate propagation is performed via a
    plate motion model (PMM) defined for the target_itrf. The target epoch
    determines the number of years into the future/past the observed points
    should be propagated. For example, if a point's observation date
    (`t_observed`) is 1993.0 (1993-01-01T00:00:00) and the target_epoch is
    2011.0 (2011-01-01T00:00:00), the point will be propagated forward 18
    years. Note that not all ITRFs have PMMs defined for them. The PMM used is
    defined for the target_epoch, so it is likely to be most accurate for points
    observed near the ITRF's defined epoch.

    All ITRF and PMM transformations are dependent on the user's `proj`
    installation's ITRF init files (see
    https://proj.org/en/9.3/resource_files.html#init-files). For example,
    ITRF2014 parameters are defined here:
    https://github.com/OSGeo/PROJ/blob/8b65d5b14e2a8fbb8198335019488a2b2968df5c/data/ITRF2014.

    Note that ILVIS2 data contain more than one set of
    latitude/longitude/elevation variables (e.g., HLAT/HLON/ZH,
    CLAT/CLON/ZC). This function only transforms the primary
    latitude/longitude/elevation fields in the provided dataframe. Use the
    `ilvis2_coordinate_set` kwarg on `read_iceflow_datafile(s)` to select an
    different primary set of latitude/longitude/elevation fields. Alternatively,
    manually set the fields:
    ```
    # TLAT/TLON/TZ are only available in ILVIS2v2 data:
    sel_ilvis2v2 = data.dataset == "ILVIS2v2"
    data.loc[sel_ilvis2v2, ["latitude", "longitude", "elevation"]] = data.loc[sel_ilvis2v2, ["TLAT", "TLON", "ZT"]]
    ```
    """
    if not check_itrf(target_itrf):
        err_msg = (
            f"The provided ITRF string was not recognized: {target_itrf}."
            " ITRF strings should be in the form 'ITRFYYYY'."
        )
        raise ValueError(err_msg)

    transformed_chunks = []
    for source_itrf, chunk in data.groupby(by="ITRF"):
        source_itrf = cast(str, source_itrf)
        # If the source ITRF is the same as the target for this chunk, skip transformation.
        if source_itrf == target_itrf and target_epoch is None:
            transformed_chunks.append(chunk)
            continue

        plate_model_step = ""
        if target_epoch:
            if not plate:
                plate = plate_name(Point(chunk.longitude.mean(), chunk.latitude.mean()))
            plate_model_step = (
                # Perform coordinate propagation to the target epoch using the
                # provided plate motion model (PMM).
                # This step uses the target_itrf's init file to lookup the
                # associated plate's PMM parameters. For example, ITRF2014
                # parameters are defined here:
                # https://github.com/OSGeo/PROJ/blob/8b65d5b14e2a8fbb8198335019488a2b2968df5c/data/ITRF2014.
                # The step is inverted because proj defined `t_epoch` as the
                # "central epoch" - not the "target epoch. The transformation
                # uses a delta defined by `t_observed - t_epoch` that are
                # applied to the PMM's rate of change to propagate the point
                # into the past/future. See
                # https://proj.org/en/9.5/operations/transformations/helmert.html#mathematical-description
                # for more information.
                # For example, if a point's observation date
                # (`t_observed`) is 1993.0 (1993-01-01T00:00:00) and the t_epoch
                # is 2011.0 (2011-01-01T00:00:00), then the delta is 1993 -
                # 2011: -18. We need to invert the step so that the point is
                # propagated forward in time, from 1993 to 2011.
                f"+step +inv +init={target_itrf}:{plate} +t_epoch={target_epoch}"
            )
            if not _check_valid_proj_step(plate_model_step):
                err_msg = f"Failed to find pre-defined plate-model parameters for {target_itrf}:{plate}"
                raise RuntimeError(err_msg)

        itrf_transformation_step = _itrf_transformation_step(source_itrf, target_itrf)

        pipeline = (
            # This initializes the pipeline and declares the use of the WGS84
            # ellipsoid for all of the following steps. See
            # https://proj.org/en/9.5/operations/pipeline.html.
            f"+proj=pipeline +ellps=WGS84 "
            # Performs unit conversion from lon/lat degrees to radians.
            # TODO: This step appears to be unnecessary. Removing it does not appear to
            # affect the output. The following steps require that the
            # coordinates be geodedic, which could be radians or degrees.
            f"+step +proj=unitconvert +xy_in=deg +xy_out=rad "
            # This step explicitly sets the projection as lat/lon. It won't
            # change the coordinates, but they will be identified as geodetic,
            # which is necessary for the next steps.
            f"+step +proj=latlon "
            # Convert from lat/lon/elev geodetic coordinates to cartesian
            # coordinates, which are required for the following steps.
            # See: https://proj.org/en/9.5/operations/conversions/cart.html
            f"+step +proj=cart "
            # ITRF transformation. See above for definition.
            f"{itrf_transformation_step} "
            # See above for definition.
            f"{plate_model_step} "
            # Convert back from cartesian to lat/lon coordinates
            f"+step +inv +proj=cart "
            # Convert lon/lat from radians back to degrees.
            # TODO: remove this if the initial conversion to radians above is not needed
            f"+step +proj=unitconvert +xy_in=rad +xy_out=deg"
        )

        transformer = pyproj.Transformer.from_pipeline(pipeline)

        decimalyears = (
            chunk.reset_index().utc_datetime.apply(_datetime_to_decimal_year).to_numpy()
        )
        # TODO: Should we create a new decimalyears when doing an epoch
        # propagation since PROJ doesn't do this?

        lons, lats, elevs, _ = transformer.transform(
            chunk.longitude,
            chunk.latitude,
            chunk.elevation,
            decimalyears,
        )

        transformed_chunk = chunk.copy()
        transformed_chunk["latitude"] = lats
        transformed_chunk["longitude"] = lons
        transformed_chunk["elevation"] = elevs
        transformed_chunk["ITRF"] = target_itrf
        transformed_chunks.append(transformed_chunk)

    transformed_df = pd.concat(transformed_chunks)
    transformed_df = transformed_df.reset_index().set_index("utc_datetime")
    return IceflowDataFrame(transformed_df)



from __future__ import annotations

import datetime as dt
from pathlib import Path
from typing import cast

import h5py
import numpy as np
import numpy.typing as npt
import pandas as pd

from nsidc.iceflow.data.models import GLAH06DataFrame

# Note: the ITRF is given by the NSIDC landing page:
# https://nsidc.org/data/glah06/versions/34#anchor-data-access-tools
# The ITRF does not appear to be present in the source data or the user guide.
GLAH06_ITRF = "ITRF2008"
VERSION = "034"
# All scalar variables with single dimension 'DS_UTCTime_40'
VARIABLES = [
    ("DS_UTCTime_40", "/Data_40HZ/DS_UTCTime_40"),
    ("i_rec_ndx", "/Data_40HZ/Time/i_rec_ndx"),
    ("i_shot_count", "/Data_40HZ/Time/i_shot_count"),
    ("d_lat", "/Data_40HZ/Geolocation/d_lat"),
    ("d_lon", "/Data_40HZ/Geolocation/d_lon"),
    ("d_elev", "/Data_40HZ/Elevation_Surfaces/d_elev"),
    ("d_refRng", "/Data_40HZ/Elevation_Surfaces/d_refRng"),
    ("d_dTrop", "/Data_40HZ/Elevation_Corrections/d_dTrop"),
    ("d_satElevCorr", "/Data_40HZ/Elevation_Corrections/d_satElevCorr"),
    ("d_GmC", "/Data_40HZ/Elevation_Corrections/d_GmC"),
    ("d_wTrop", "/Data_40HZ/Elevation_Corrections/d_wTrop"),
    ("d_beamCoelv", "/Data_40HZ/Elevation_Angles/d_beamCoelv"),
    ("d_beamAzimuth", "/Data_40HZ/Elevation_Angles/d_beamAzimuth"),
    ("d_SigBegOff", "/Data_40HZ/Elevation_Offsets/d_SigBegOff"),
    ("d_TrshRngOff", "/Data_40HZ/Elevation_Offsets/d_TrshRngOff"),
    ("d_SigEndOff", "/Data_40HZ/Elevation_Offsets/d_SigEndOff"),
    ("d_cntRngOff", "/Data_40HZ/Elevation_Offsets/d_cntRngOff"),
    ("d_isRngOff", "/Data_40HZ/Elevation_Offsets/d_isRngOff"),
    ("d_siRngOff", "/Data_40HZ/Elevation_Offsets/d_siRngOff"),
    ("d_ldRngOff", "/Data_40HZ/Elevation_Offsets/d_ldRngOff"),
    ("d_ocRngOff", "/Data_40HZ/Elevation_Offsets/d_ocRngOff"),
    ("rng_uqf_sigbeg1_flg", "/Data_40HZ/Quality/rng_uqf_sigbeg1_flg"),
    ("rng_uqf_sigend1_flg", "/Data_40HZ/Quality/rng_uqf_sigend1_flg"),
    ("rng_uqf_thres1_flg", "/Data_40HZ/Quality/rng_uqf_thres1_flg"),
    ("rng_uqf_cent1_flg", "/Data_40HZ/Quality/rng_uqf_cent1_flg"),
    ("rng_uqf_sigbeg2_flg", "/Data_40HZ/Quality/rng_uqf_sigbeg2_flg"),
    ("rng_uqf_sigend2_flg", "/Data_40HZ/Quality/rng_uqf_sigend2_flg"),
    ("rng_uqf_thres2_flg", "/Data_40HZ/Quality/rng_uqf_thres2_flg"),
    ("rng_uqf_cent2_flg", "/Data_40HZ/Quality/rng_uqf_cent2_flg"),
    ("rng_uqf_is_flg", "/Data_40HZ/Quality/rng_uqf_is_flg"),
    ("rng_uqf_si_flg", "/Data_40HZ/Quality/rng_uqf_si_flg"),
    ("rng_uqf_ld_flg", "/Data_40HZ/Quality/rng_uqf_ld_flg"),
    ("rng_uqf_oc_flg", "/Data_40HZ/Quality/rng_uqf_oc_flg"),
    ("sat_corr_flg", "/Data_40HZ/Quality/sat_corr_flg"),
    ("elev_use_flg", "/Data_40HZ/Quality/elev_use_flg"),
    ("att_pad_use_flg", "/Data_40HZ/Quality/att_pad_use_flg"),
    ("att_calc_pad_flg", "/Data_40HZ/Quality/att_calc_pad_flg"),
    ("att_lpa_flg", "/Data_40HZ/Quality/att_lpa_flg"),
    ("sigma_att_flg", "/Data_40HZ/Quality/sigma_att_flg"),
    ("i_satNdx", "/Data_40HZ/Quality/i_satNdx"),
    ("d_pctSAT", "/Data_40HZ/Quality/d_pctSAT"),
    ("elv_cnt_1_flg", "/Data_40HZ/Elevation_Flags/elv_cnt_1_flg"),
    ("elv_cnt_2_flg", "/Data_40HZ/Elevation_Flags/elv_cnt_2_flg"),
    ("elv_peak_1_flg", "/Data_40HZ/Elevation_Flags/elv_peak_1_flg"),
    ("elv_peak_2_flg", "/Data_40HZ/Elevation_Flags/elv_peak_2_flg"),
    ("elv_thres_flg", "/Data_40HZ/Elevation_Flags/elv_thres_flg"),
    ("elv_gauss_flg", "/Data_40HZ/Elevation_Flags/elv_gauss_flg"),
    ("elv_other_flg", "/Data_40HZ/Elevation_Flags/elv_other_flg"),
    ("elv_cloud_flg", "/Data_40HZ/Elevation_Flags/elv_cloud_flg"),
    ("d_TxNrg", "/Data_40HZ/Transmit_Energy/d_TxNrg"),
    ("d_d2refTrk", "/Data_40HZ/Geophysical/d_d2refTrk"),
    ("d_DEM_elv", "/Data_40HZ/Geophysical/d_DEM_elv"),
    ("d_ocElv", "/Data_40HZ/Geophysical/d_ocElv"),
    ("d_poTide", "/Data_40HZ/Geophysical/d_poTide"),
    ("d_gdHt", "/Data_40HZ/Geophysical/d_gdHt"),
    ("d_erElv", "/Data_40HZ/Geophysical/d_erElv"),
    ("d_eqElv", "/Data_40HZ/Geophysical/d_eqElv"),
    ("d_ldElv", "/Data_40HZ/Geophysical/d_ldElv"),
    ("d_deltaEllip", "/Data_40HZ/Geophysical/d_deltaEllip"),
    ("d_ElevBiasCorr", "/Data_40HZ/Geophysical/d_ElevBiasCorr"),
    ("i_DEM_hires_src_1", "/Data_40HZ/Geophysical/i_DEM_hires_src_1"),
    ("d_reflctUC", "/Data_40HZ/Reflectivity/d_reflctUC"),
    ("d_sDevNsOb1", "/Data_40HZ/Reflectivity/d_sDevNsOb1"),
    ("d_satNrgCorr", "/Data_40HZ/Reflectivity/d_satNrgCorr"),
    ("d_RecNrgAll", "/Data_40HZ/Reflectivity/d_RecNrgAll"),
    ("d_skew2", "/Data_40HZ/Waveform/d_skew2"),
    ("d_kurt2", "/Data_40HZ/Waveform/d_kurt2"),
    ("d_maxRecAmp", "/Data_40HZ/Waveform/d_maxRecAmp"),
    ("d_maxSmAmp", "/Data_40HZ/Waveform/d_maxSmAmp"),
    ("i_nPeaks1", "/Data_40HZ/Waveform/i_nPeaks1"),
    ("i_numPk", "/Data_40HZ/Waveform/i_numPk"),
    ("i_gval_rcv", "/Data_40HZ/Waveform/i_gval_rcv"),
    ("d_FRir_cldtop", "/Data_40HZ/Atmosphere/d_FRir_cldtop"),
    ("FRir_qa_flg", "/Data_40HZ/Atmosphere/FRir_qa_flg"),
    ("d_FRir_intsig", "/Data_40HZ/Atmosphere/d_FRir_intsig"),
]

TIMESTAMP_COLUMN = "utc_datetime"

DATA_COLUMNS = [
    ("i_rec_ndx", b"%d"),
    ("i_shot_count", b"%d"),
    ("d_lat", b"%f"),
    ("d_lon", b"%f"),
    ("d_elev", b"%f"),
    ("d_refRng", b"%f"),
    ("d_dTrop", b"%f"),
    ("d_satElevCorr", b"%f"),
    ("d_GmC", b"%f"),
    ("d_wTrop", b"%f"),
    ("d_beamCoelv", b"%f"),
    ("d_beamAzimuth", b"%f"),
    ("d_SigBegOff", b"%f"),
    ("d_TrshRngOff", b"%f"),
    ("d_SigEndOff", b"%f"),
    ("d_cntRngOff", b"%f"),
    ("d_isRngOff", b"%f"),
    ("d_siRngOff", b"%f"),
    ("d_ldRngOff", b"%f"),
    ("d_ocRngOff", b"%f"),
    ("rng_uqf_sigbeg1_flg", b"%d"),
    ("rng_uqf_sigend1_flg", b"%d"),
    ("rng_uqf_thres1_flg", b"%d"),
    ("rng_uqf_cent1_flg", b"%d"),
    ("rng_uqf_sigbeg2_flg", b"%d"),
    ("rng_uqf_sigend2_flg", b"%d"),
    ("rng_uqf_thres2_flg", b"%d"),
    ("rng_uqf_cent2_flg", b"%d"),
    ("rng_uqf_is_flg", b"%d"),
    ("rng_uqf_si_flg", b"%d"),
    ("rng_uqf_ld_flg", b"%d"),
    ("rng_uqf_oc_flg", b"%d"),
    ("sat_corr_flg", b"%d"),
    ("elev_use_flg", b"%d"),
    ("att_pad_use_flg", b"%d"),
    ("att_calc_pad_flg", b"%d"),
    ("att_lpa_flg", b"%d"),
    ("sigma_att_flg", b"%d"),
    ("i_satNdx", b"%d"),
    ("d_pctSAT", b"%f"),
    ("elv_cnt_1_flg", b"%d"),
    ("elv_cnt_2_flg", b"%d"),
    ("elv_peak_1_flg", b"%d"),
    ("elv_peak_2_flg", b"%d"),
    ("elv_thres_flg", b"%d"),
    ("elv_gauss_flg", b"%d"),
    ("elv_other_flg", b"%d"),
    ("elv_cloud_flg", b"%d"),
    ("d_TxNrg", b"%f"),
    ("d_d2refTrk", b"%f"),
    ("d_DEM_elv", b"%f"),
    ("d_ocElv", b"%f"),
    ("d_poTide", b"%f"),
    ("d_gdHt", b"%f"),
    ("d_erElv", b"%f"),
    ("d_eqElv", b"%f"),
    ("d_ldElv", b"%f"),
    ("d_deltaEllip", b"%f"),
    ("d_ElevBiasCorr", b"%f"),
    ("i_DEM_hires_src_1", b"%d"),
    ("d_reflctUC", b"%f"),
    ("d_sDevNsOb1", b"%f"),
    ("d_satNrgCorr", b"%f"),
    ("d_RecNrgAll", b"%f"),
    ("d_skew2", b"%f"),
    ("d_kurt2", b"%f"),
    ("d_maxRecAmp", b"%f"),
    ("d_maxSmAmp", b"%f"),
    ("i_nPeaks1", b"%d"),
    ("i_numPk", b"%d"),
    ("i_gval_rcv", b"%d"),
    ("d_FRir_cldtop", b"%f"),
    ("FRir_qa_flg", b"%d"),
    ("d_FRir_intsig", b"%f"),
]


def _utc_datetime(seconds):
    """Return 'utc_datetime' Series with values calculated from the shot data.

    The transmit time of each shot in the 1 second frame is measured as UTC seconds elapsed
    since Jan 1 2000 12:00:00 UTC. This time has been derived from the GPS time accounting
    for leap seconds.
    """
    epoc = dt.datetime(2000, 1, 1, 12, 0, 0)
    utc = seconds.apply(lambda s: epoc + dt.timedelta(seconds=s))

    return utc


def _mask_invalid(var) -> npt.NDArray[np.bool_]:
    assert len(var.shape) == 1, "Expected only 1 dimensional data"
    values = var[:]

    invalid: npt.NDArray[np.bool_] = np.full(var.shape, False)

    # Mask no data values
    if "_FillValue" in var.attrs:
        invalid |= values == var.attrs["_FillValue"][0]

    # Mask out-of-range values
    if "valid_min" in var.attrs and "valid_max" in var.attrs:
        valid_min = var.attrs["valid_min"][0]
        valid_max = var.attrs["valid_max"][0]
        invalid |= (values < valid_min) | (values > valid_max)

    return invalid


def _glah06_dataframe(filepath):
    """Returns a GLAH06 DataFrame read from an HDF5 file."""
    df = pd.DataFrame()
    with h5py.File(filepath, "r") as glah06:
        invalid_geo = None
        for name, path in VARIABLES:
            var = glah06[path]
            df[name] = var[:]

            # Mask row completely if geolocation is missing
            if name in ("d_lat", "d_lon", "d_elev"):
                if invalid_geo is None:
                    invalid_geo = np.full(var.shape, False)

                invalid_geo |= _mask_invalid(var)

    invalid_geo = cast(npt.NDArray[np.bool_], invalid_geo)
    df = df[~invalid_geo]

    return df


def _glah06_data(filepath):
    df = _glah06_dataframe(filepath)

    df["utc_datetime"] = _utc_datetime(df["DS_UTCTime_40"]) if not df.empty else None

    df = df.drop(columns="DS_UTCTime_40")

    return df


def glah06_data(filepath: Path) -> GLAH06DataFrame:
    """Return an GLAH06 file DataFrame, performing all necessary
    conversions / augmentation on the data.
    """
    df = _glah06_data(filepath)

    # Add the ITRF
    df["ITRF"] = GLAH06_ITRF
    # To be consistent with the other `nsidc-iceflow` datasets, copy the primary
    # lat/lon/elev fields to the standard "latitude", "longitude", "elevation"
    # field names.
    df["latitude"] = df["d_lat"]
    df["longitude"] = df["d_lon"]
    df["elevation"] = df["d_elev"]
    df["dataset"] = "GLAH06v034"

    # We index the data by utc datetime.
    df = df.set_index("utc_datetime")

    return GLAH06DataFrame(df)



"""End-to-End test for the typical iceflow pipeline.

* Searches for small sample of data
* Downloads small sample of data
* Performs ITRF transformation

This serves as prototype for planned Jupyter Notebook-based tutorial featuring
this library.
"""

from __future__ import annotations

import datetime as dt
from pathlib import Path

import dask.dataframe as dd
import pandas as pd

from nsidc.iceflow import (
    download_iceflow_results,
    find_iceflow_data,
    make_iceflow_parquet,
)
from nsidc.iceflow.data.ilvis2 import (
    ILVIS2_COORDINATE_SETS,
    ILVIS2_DEFAULT_COORDINATE_SET,
)
from nsidc.iceflow.data.models import (
    BoundingBoxLike,
    Dataset,
    IceflowDataFrame,
    TemporalRange,
)
from nsidc.iceflow.data.read import read_iceflow_datafiles
from nsidc.iceflow.data.supported_datasets import (
    ALL_SUPPORTED_DATASETS,
    BLATM1BDataset,
    GLAH06Dataset,
    ILATM1BDataset,
    ILVIS2Dataset,
)
from nsidc.iceflow.itrf.converter import transform_itrf


def _fetch_iceflow_df(
    *,
    bounding_box: BoundingBoxLike,
    temporal: TemporalRange,
    datasets: list[Dataset] = ALL_SUPPORTED_DATASETS,
    output_dir: Path,
    output_itrf: str | None = None,
    ilvis2_coordinate_set: ILVIS2_COORDINATE_SETS = ILVIS2_DEFAULT_COORDINATE_SET,
):
    """Search for data matching parameters and return an IceflowDataframe.

    Optionally transform data to the given ITRF for consistency.

    Note: a potentially large amount of data may be returned, especially if the
    user requests a large spatial/temporal area across multiple datasets. The
    result may not even fit in memory!

    Consider using `make_iceflow_parquet` to store downloaded data in parquet
    format.
    """

    iceflow_search_reuslts = find_iceflow_data(
        bounding_box=bounding_box,
        temporal=temporal,
        datasets=datasets,
    )

    downloaded_files = download_iceflow_results(
        iceflow_search_results=iceflow_search_reuslts,
        output_dir=output_dir,
    )

    iceflow_df = read_iceflow_datafiles(
        downloaded_files, ilvis2_coordinate_set=ilvis2_coordinate_set
    )

    if output_itrf is not None:
        iceflow_df = transform_itrf(
            data=iceflow_df,
            target_itrf=output_itrf,
        )

    return iceflow_df


def test_atm1b_ilatm1b(tmp_path):
    target_itrf = "ITRF2014"
    common_bounding_box = (
        -103.125559,
        -75.180563,
        -102.677327,
        -74.798063,
    )

    # Native ITRF is ITRF2005
    results_ilatm1b_v1_2009 = _fetch_iceflow_df(
        datasets=[ILATM1BDataset(version="1")],
        bounding_box=common_bounding_box,
        temporal=(dt.date(2009, 11, 1), dt.date(2009, 12, 1)),
        output_dir=tmp_path,
        output_itrf=target_itrf,
    )

    assert (results_ilatm1b_v1_2009.dataset == "ILATM1Bv1").all()

    # Native ITRF is ITRF2008
    results_ilatm1b_v2_2014 = _fetch_iceflow_df(
        datasets=[ILATM1BDataset(version="2")],
        bounding_box=common_bounding_box,
        temporal=(dt.date(2014, 11, 1), dt.date(2014, 12, 1)),
        output_dir=tmp_path,
        output_itrf=target_itrf,
    )

    assert (results_ilatm1b_v2_2014.dataset == "ILATM1Bv2").all()

    complete_df = IceflowDataFrame(
        pd.concat([results_ilatm1b_v1_2009, results_ilatm1b_v2_2014])
    )

    assert (complete_df.ITRF.unique() == target_itrf).all()


def test_atm1b_blatm1b(tmp_path):
    common_bounding_box = (
        -120.0,
        -75.1,
        -92.0,
        -65.0,
    )

    results_blamt1b_v1_2014 = _fetch_iceflow_df(
        datasets=[BLATM1BDataset()],
        bounding_box=common_bounding_box,
        temporal=(dt.date(2002, 11, 27), dt.date(2002, 11, 28)),
        output_dir=tmp_path,
    )

    assert (results_blamt1b_v1_2014.dataset == "BLATM1Bv1").all()

    assert (results_blamt1b_v1_2014.ITRF == "ITRF2000").all()


def test_ivlis2(tmp_path):
    results_v1 = _fetch_iceflow_df(
        datasets=[ILVIS2Dataset(version="1")],
        bounding_box=(
            -120.0,
            -80.0,
            -90.0,
            -65.0,
        ),
        temporal=(dt.datetime(2009, 10, 25, 15), dt.datetime(2009, 10, 25, 17)),
        output_dir=tmp_path,
    )

    assert (results_v1.dataset == "ILVIS2v1").all()

    assert (results_v1.ITRF == "ITRF2000").all()

    results_v2 = _fetch_iceflow_df(
        datasets=[ILVIS2Dataset(version="2")],
        bounding_box=(
            -180,
            60.0,
            180,
            90,
        ),
        temporal=(dt.datetime(2017, 8, 25, 0), dt.datetime(2017, 8, 25, 14, 30)),
        output_dir=tmp_path,
    )

    assert (results_v2.dataset == "ILVIS2v2").all()

    assert (results_v2.ITRF == "ITRF2008").all()

    # test that v1 and 2 can be concatenated
    complete_df = IceflowDataFrame(pd.concat([results_v1, results_v2]))

    assert complete_df is not None

    # Confirm that we get the alt. set of ilvis2 coordinates when specified.
    results_v2_alt_coords = _fetch_iceflow_df(
        datasets=[ILVIS2Dataset(version="2")],
        bounding_box=(
            -180,
            60.0,
            180,
            90,
        ),
        temporal=(dt.datetime(2017, 8, 25, 0), dt.datetime(2017, 8, 25, 14, 30)),
        output_dir=tmp_path,
        ilvis2_coordinate_set="highest_signal",
    )
    assert (results_v2_alt_coords.latitude == results_v2_alt_coords.TLAT).all()
    assert (results_v2_alt_coords.longitude == results_v2_alt_coords.TLON).all()
    assert (results_v2_alt_coords.elevation == results_v2_alt_coords.ZT).all()

    # Results v1 should use the default ("low_mode")
    assert (results_v2.latitude == results_v2.GLAT).all()
    assert (results_v2.longitude == results_v2.GLON).all()
    assert (results_v2.elevation == results_v2.ZG).all()


def test_glah06(tmp_path):
    common_bounding_box = (
        -180,
        -90,
        180,
        90,
    )

    results = _fetch_iceflow_df(
        datasets=[GLAH06Dataset()],
        bounding_box=common_bounding_box,
        temporal=(
            dt.datetime(2003, 2, 20, 22, 25),
            dt.datetime(2003, 2, 20, 22, 25, 38),
        ),
        output_dir=tmp_path,
    )

    assert (results.dataset == "GLAH06v034").all()
    assert (results.ITRF == "ITRF2008").all()


def _create_iceflow_parquet(
    *,
    bounding_box: BoundingBoxLike,
    temporal: TemporalRange,
    datasets: list[Dataset] = ALL_SUPPORTED_DATASETS,
    output_dir: Path,
    target_itrf: str,
    overwrite: bool = False,
    target_epoch: str | None = None,
) -> Path:
    """Create a parquet dataset containing the lat/lon/elev data matching the dataset search params.

    This function creates a parquet dataset that can be easily used alongside dask,
    containing lat/lon/elev data.

    Note: this function writes a single `iceflow.parquet` to the output
    dir. This code does not currently support updates to the parquet after being
    written. This is intended to help facilitate analysis of a specific area
    over time. If an existing `iceflow.parquet` exists and the user wants to
    create a new `iceflow.parquet` for a different area or timespan, they will
    need to move/remove the existing `iceflow.parquet` first (e.g., with the
    `overwrite=True` kwarg).
    """
    iceflow_search_results = find_iceflow_data(
        datasets=datasets,
        temporal=temporal,
        bounding_box=bounding_box,
    )

    download_iceflow_results(
        iceflow_search_results=iceflow_search_results,
        output_dir=output_dir,
    )

    parquet_path = make_iceflow_parquet(
        data_dir=output_dir,
        target_itrf=target_itrf,
        overwrite=overwrite,
        target_epoch=target_epoch,
    )

    return parquet_path


def test_create_iceflow_parquet(tmp_path):
    target_itrf = "ITRF2014"
    common_bounding_box = (
        -49.149,
        69.186,
        -48.949,
        69.238,
    )

    # This should finds 4 results for ILATM1B v1 and 3 results for v2.
    parquet_path = _create_iceflow_parquet(
        datasets=[ILATM1BDataset(version="1"), ILATM1BDataset(version="2")],
        bounding_box=common_bounding_box,
        temporal=((dt.date(2007, 1, 1), dt.date(2014, 10, 28))),
        output_dir=tmp_path,
        target_itrf=target_itrf,
    )

    df = dd.read_parquet(parquet_path)

    # Assert that the parquet data has the expected columns
    expected_columns = sorted(["latitude", "longitude", "elevation", "dataset"])
    assert expected_columns == sorted(df.columns)

    # Assert that the two datasets we expect are present.
    assert sorted(["ILATM1Bv1", "ILATM1Bv2"]) == sorted(
        df.dataset.unique().compute().values
    )



from __future__ import annotations

from pathlib import Path

import pandas as pd

from nsidc.iceflow.data.atm1b import atm1b_data
from nsidc.iceflow.data.glah06 import glah06_data
from nsidc.iceflow.data.ilvis2 import (
    ILVIS2_COORDINATE_SETS,
    ILVIS2_DEFAULT_COORDINATE_SET,
    ilvis2_data,
)
from nsidc.iceflow.data.models import (
    ATM1BDataFrame,
    GLAH06DataFrame,
    IceflowDataFrame,
    ILVIS2DataFrame,
)


def read_iceflow_datafile(
    filepath: Path,
    ilvis2_coordinate_set: ILVIS2_COORDINATE_SETS = ILVIS2_DEFAULT_COORDINATE_SET,
) -> IceflowDataFrame | ATM1BDataFrame | ILVIS2DataFrame | GLAH06DataFrame:
    """Read the given iceflow datafile."""
    # iceflow data are expected to exist in a directory named like
    # `{short_name}_{version}`
    dataset_subdir = filepath.parent.name
    short_name, _version = dataset_subdir.split("_")

    if short_name in ["ILATM1B", "BLATM1B"]:
        return atm1b_data(filepath)
    elif short_name == "ILVIS2":
        return ilvis2_data(filepath, coordinate_set=ilvis2_coordinate_set)
    elif short_name == "GLAH06":
        return glah06_data(filepath)
    else:
        err_msg = f"Unrecognized dataset {short_name=} extracted from {filepath.parent}"
        raise RuntimeError(err_msg)


def read_iceflow_datafiles(
    filepaths: list[Path],
    ilvis2_coordinate_set: ILVIS2_COORDINATE_SETS = ILVIS2_DEFAULT_COORDINATE_SET,
) -> IceflowDataFrame:
    all_dfs = []
    for filepath in filepaths:
        df = read_iceflow_datafile(
            filepath,
            ilvis2_coordinate_set=ilvis2_coordinate_set,
        )
        all_dfs.append(df)

    complete_df = IceflowDataFrame(pd.concat(all_dfs))

    return complete_df



from __future__ import annotations

import pandas as pd
import pandera as pa
import pytest

from nsidc.iceflow.data.models import IceflowDataFrame

_mock_bad_df = pd.DataFrame(
    {
        "latitude": [70],
        "longitude": [-50],
        "elevation": [1],
        "ITRF": ["should fail"],
    },
)


def test_iceflowdataframe():
    with pytest.raises(pa.errors.SchemaError):
        IceflowDataFrame(_mock_bad_df)


@pa.check_types
def _pa_check_in(_df: IceflowDataFrame) -> None:
    return None


def test_pa_check_in():
    with pytest.raises(pa.errors.SchemaError):
        # The type ignore on the next line tells mypy to ignore that we're not
        # casting to the expected input type. We want to test that
        # `check_types` raises a runtime validation error
        _pa_check_in(_mock_bad_df)  # type: ignore[arg-type]


@pa.check_types
def _pa_check_out(df: pd.DataFrame) -> IceflowDataFrame:
    # The type ignore on the next line tells mypy to ignore that we're not
    # casting to the expected return type. We want to test that
    # `check_types` raises a runtime validation error
    return df  # type: ignore[return-value]


def test_pa_check_out():
    with pytest.raises(pa.errors.SchemaError):
        _pa_check_out(_mock_bad_df)
